from astroid import nodes, Context  # type: ignore
from typing import TYPE_CHECKING, Optional, List, Union, Tuple
from enum import Enum, auto
from functools import reduce

from pylint.checkers import BaseChecker  # type: ignore
from pylint.checkers.utils import only_required_for_messages

if TYPE_CHECKING:
    from pylint.lint import PyLinter  # type: ignore

from edulint.linting.checkers.utils import (
    get_name,
    is_builtin,
    get_range_params,
    get_statements_count,
    is_parents_elif,
)
from edulint.linting.checkers.modified_listener import ModifiedListener


class UsesIndex(Enum):
    OUTSIDE_SUBSCRIPT = auto()
    INSIDE_SUBSCRIPT = auto()
    NEVER = auto()

    @classmethod
    def combine(cls, lt: "UsesIndex", rt: "UsesIndex") -> "UsesIndex":
        if lt == cls.INSIDE_SUBSCRIPT or rt == cls.INSIDE_SUBSCRIPT:
            return cls.INSIDE_SUBSCRIPT
        if lt == cls.OUTSIDE_SUBSCRIPT or rt == cls.OUTSIDE_SUBSCRIPT:
            return cls.OUTSIDE_SUBSCRIPT
        return cls.NEVER


class ImproveForLoop(BaseChecker):  # type: ignore
    name = "improve-for-loop"
    msgs = {
        "R6101": (
            'Iterate directly: "for var in %s" (with appropriate name for "var")',
            "use-foreach",
            "Emitted when a for-range loop is used while a for-each loop would suffice.",
        ),
        "R6102": (
            'Iterate using enumerate: "for %s, var in enumerate(%s)" (with appropriate name for "var")',
            "use-enumerate",
            "Emitted when a for-range loop is used with the element at each index is accessed as well.",
        ),
    }

    def __init__(self, linter: Optional["PyLinter"] = None) -> None:
        super().__init__(linter)

    class StructureIndexedVisitor(ModifiedListener[Tuple[bool, bool]]):
        default = (False, False)

        @staticmethod
        def combine(results: List[Tuple[bool, bool]]) -> Tuple[bool, bool]:
            return any(loaded for loaded, _stored in results), any(
                stored for _loaded, stored in results
            )

        def __init__(self, structure: Union[nodes.Name, nodes.Attribute], index: nodes.Name):
            super().__init__([structure, index])
            self.structure = structure
            self.index = index

        def visit_subscript(self, subscript: nodes.Subscript) -> Tuple[bool, bool]:
            sub_loaded, sub_stored = self.visit_many(subscript.get_children())

            if self.was_reassigned(self.structure, allow_definition=False) or self.was_reassigned(
                self.index, allow_definition=False
            ):
                return False, False

            used = (
                subscript.value.as_string() == self.structure.as_string()
                and isinstance(subscript.slice, nodes.Name)
                and subscript.slice.as_string() == self.index.as_string()
            )

            if subscript.ctx == Context.Store:
                return sub_loaded, sub_stored or used

            if subscript.ctx == Context.Load:
                return sub_loaded or used, sub_stored

            assert subscript.ctx == Context.Del
            return sub_loaded, sub_stored or used

    class IndexUsedVisitor(ModifiedListener[UsesIndex]):
        default = UsesIndex.NEVER

        @staticmethod
        def combine(results: List[UsesIndex]) -> UsesIndex:
            return reduce(UsesIndex.combine, results, UsesIndex.NEVER)

        def __init__(self, structure: Union[nodes.Name, nodes.Attribute], index: nodes.Name):
            super().__init__([structure, index])
            self.structure = structure
            self.index = index

        def visit_name(self, name: nodes.Name) -> UsesIndex:
            if name.name != self.index.name:
                return UsesIndex.NEVER

            parent = name.parent
            if (
                isinstance(parent, nodes.Subscript)
                and parent.value.as_string() != self.structure.as_string()
            ):
                return UsesIndex.INSIDE_SUBSCRIPT

            if (
                not isinstance(name.parent, nodes.Subscript)
                or not isinstance(self.structure, type(name.parent.value))
                or self.was_reassigned(name, allow_definition=False)
            ):
                return UsesIndex.OUTSIDE_SUBSCRIPT

            subscript = name.parent
            if not (
                (
                    isinstance(subscript.value, nodes.Name)
                    and subscript.value.name == self.structure.name
                )
                or (
                    isinstance(subscript.value, nodes.Attribute)
                    and subscript.value.attrname == self.structure.attrname
                )
            ):
                return UsesIndex.OUTSIDE_SUBSCRIPT

            if isinstance(subscript.parent, nodes.Assign) and subscript in subscript.parent.targets:
                return UsesIndex.OUTSIDE_SUBSCRIPT

            return UsesIndex.NEVER

    def visit_for(self, node: nodes.For) -> None:
        range_params = get_range_params(node.iter)
        if range_params is None:
            return

        start, stop, step = range_params
        if (
            not isinstance(start, nodes.Const)
            or start.value != 0
            or not isinstance(stop, nodes.Call)
            or not is_builtin(stop.func, "len")
            or len(stop.args) != 1
            or not isinstance(step, nodes.Const)
            or step.value != 1
        ):
            return

        structure = stop.args[0]
        index = node.target
        if not isinstance(structure, nodes.Name) and not isinstance(structure, nodes.Attribute):
            return

        loaded, stored = self.StructureIndexedVisitor(structure, node.target).visit_many(node.body)
        if not loaded:
            return

        structure_name = get_name(structure)
        uses_index = self.IndexUsedVisitor(structure, node.target).visit_many(node.body)

        if uses_index == UsesIndex.INSIDE_SUBSCRIPT:
            return
        elif stored or uses_index == UsesIndex.OUTSIDE_SUBSCRIPT:
            self.add_message("use-enumerate", args=(index.name, structure_name), node=node)
        else:
            assert uses_index == UsesIndex.NEVER
            self.add_message("use-foreach", args=structure_name, node=node)


class NoGlobalVars(BaseChecker):
    name = "no-global-variables"
    msgs = {
        "R6401": (
            "Do not use global variables; you use %s, modifying it for example at line %i.",
            "no-global-vars",
            "Emitted when the code uses global variables.",
        ),
    }

    def __init__(self, linter: "PyLinter"):
        super().__init__(linter)
        self.to_check = {}

    def visit_assignname(self, node: nodes.AssignName) -> None:
        frame = node.frame()
        if not isinstance(frame, nodes.Module):
            return

        if frame not in self.to_check:
            self.to_check[frame] = {}

        if node.name in self.to_check[frame].keys():
            return

        self.to_check[frame][node.name] = node

    def close(self) -> None:
        for frame, vars_ in self.to_check.items():
            listener = ModifiedListener(list(vars_.values()))
            listener.visit(frame)
            for node in vars_.values():
                if listener.was_modified(node, allow_definition=True):
                    nonglobal_modifiers = [
                        n for n in listener.get_all_modifiers(node) if n.scope() != node.scope()
                    ]
                    if nonglobal_modifiers:
                        self.add_message(
                            "no-global-vars",
                            node=node,
                            args=(node.name, nonglobal_modifiers[0].lineno),
                        )


class LongCodeChecker(BaseChecker):
    name = "long-code"
    msgs = {
        "R6701": (
            "Too much code outside of functions or classes (%d which is over %d statements).",
            "long-script",
            "Emitted when there are too many lines of code on the top level that are not import or function or class "
            "definition.",
        ),
        "R6702": (
            "Function '%s' is too long (%d which is over %d statements).",
            "long-function",
            "Emitted when there are too many statements inside a function definition.",
        ),
        "R6703": (
            "Use early return.",
            "use-early-return",
            "Emitted when a long block of code is followed by an else that just returns, breaks or continues.",
        ),
    }

    @only_required_for_messages("long-script")
    def visit_module(self, node: nodes.Module):
        MAX_SCRIPT = 20

        count = get_statements_count(node, include_defs=False, include_name_main=False)
        if count > MAX_SCRIPT:
            self.add_message("long-script", node=node, args=(count, MAX_SCRIPT))

    @only_required_for_messages("long-function")
    def visit_functiondef(self, node: nodes.FunctionDef):
        MAX_FUNC = 20

        count = get_statements_count(node.body, include_defs=False, include_name_main=False)
        if count > MAX_FUNC:
            self.add_message("long-function", node=node, args=(node.name, count, MAX_FUNC))

    @only_required_for_messages("use-early-return")
    def visit_if(self, node: nodes.If):
        def ends_block(node: nodes.NodeNG) -> bool:
            if isinstance(node, nodes.If):
                return (
                    ends_block(node.body[-1])
                    and len(node.orelse) > 0
                    and ends_block(node.orelse[-1])
                )
            return isinstance(node, (nodes.Return, nodes.Break, nodes.Continue))

        if is_parents_elif(node):
            return

        if len(node.orelse) > 0:
            if get_statements_count(node.orelse, include_defs=True, include_name_main=True) > 2:
                return
            last = node.orelse[-1]
        elif ends_block(node.body[-1]):
            last = node.next_sibling()
        else:
            return

        if (
            ends_block(last)
            and get_statements_count(node.body, include_defs=True, include_name_main=True) > 3
        ):
            self.add_message("use-early-return", node=node)


def register(linter: "PyLinter") -> None:
    """This required method auto registers the checker during initialization.
    :param linter: The linter to register the checker to.
    """
    linter.register_checker(ImproveForLoop(linter))
    linter.register_checker(NoGlobalVars(linter))
    linter.register_checker(LongCodeChecker(linter))



import pytest
from edulint.options import Option
from edulint.config.arg import Arg
from edulint.linting.problem import Problem
from utils import lazy_problem, apply_and_lint
from typing import List


@pytest.mark.parametrize("filename,expected_output", [
    ("013165-coins.py", [
        lazy_problem().set_line(9)
        .set_text("Identical code inside all if's branches, move 1 lines before the if.")
    ]),
    ("015726-prime.py", [
        # lazy_problem().set_line(10)
        # .set_text("Identical code inside all if's branches, move 1 lines after the if.")
    ]),
    ("016119-fibsum.py", [
        lazy_problem().set_line(22)
        .set_text("Identical code inside all if's branches, move 3 lines after the if.")
    ]),
    ("023171-credit.py", [
        lazy_problem().set_line(26),
        lazy_problem().set_line(38)
    ]),
    ("024056-cards.py", [
        lazy_problem().set_line(7)
    ]),
    ("024329-workdays.py", [  # TODO improve
        # lazy_problem().set_line(56)
    ]),
    ("024535-credit.py", [
        lazy_problem().set_line(23)
        .set_text("Identical code inside all if's branches, move 3 lines after the if."),
        lazy_problem().set_line(35),
    ]),
    ("052786-course.py", []),
    ("054050-course.py", [
        lazy_problem().set_line(31)
    ]),
    ("074168-doubly_linked.py", []),
    ("074242-tortoise.py", [
        lazy_problem().set_line(32),
        lazy_problem().set_line(32),
        lazy_problem().set_line(65),
        lazy_problem().set_line(65)
    ]),
    ("124573-lists.py", [
        lazy_problem().set_line(51)
        .set_text("Identical code inside all if's branches, move 1 lines after the if.")
    ]),
    ("hw21739.py", [  # TODO more precise detection
        lazy_problem().set_line(40)
        .set_text("Identical code inside all if's branches, move 9 lines after the if."),
        lazy_problem().set_line(42)
        .set_text("Identical code inside all if's branches, move 1 lines after the if."),
        lazy_problem().set_line(53)
        .set_text("Identical code inside all if's branches, move 1 lines after the if."),
        # lazy_problem().set_line(67)
        # .set_text("Identical code inside all if's branches, move 1 lines after the if."),
        # lazy_problem().set_line(87)
        # .set_text("Identical code inside all if's branches, move 2 lines before the if."),
        # lazy_problem().set_line(87)
        # .set_text("Identical code inside all if's branches, move 3 lines after the if."),
        lazy_problem().set_line(109)
        .set_text("Identical code inside all if's branches, move 16 lines after the if."),
        lazy_problem().set_line(164)
        .set_text("Identical code inside all if's branches, move 2 lines before the if."),
        lazy_problem().set_line(164)
        .set_text("Identical code inside all if's branches, move 3 lines after the if.")
    ]),
    ("hw48505.py", []),
])
def test_duplicate_if_branches(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=duplicate-if-branches")],
        expected_output
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("012889-geometry.py", [
        lazy_problem().set_line(28).set_end_line(33)
        .set_text("Identical code inside 3 consecutive ifs, join their conditions using 'or'.")
    ]),
    ("012889-pythagorean.py", [
        lazy_problem().set_line(26).set_end_line(29)
        .set_text("Identical code inside 2 consecutive ifs, join their conditions using 'or'.")
    ]),
    ("014341-geometry.py", [
        lazy_problem().set_line(9).set_end_line(14)
    ]),
    ("014613-next.py", [
        lazy_problem().set_line(35).set_end_line(40)
    ]),
    ("034440-cellular.py", [
        lazy_problem().set_line(56).set_end_line(61),
        lazy_problem().set_line(62).set_end_line(65)
    ]),
    ("046586-person_id.py", [
        lazy_problem().set_line(40).set_end_line(56)
        .set_text("Identical code inside 7 consecutive ifs, join their conditions using 'or'.")
    ]),
    ("052975-parse_time.py", []),
])
def test_duplicate_seq_ifs(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=duplicate-seq-ifs")],
        expected_output
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("012853-geometry.py", [  # TODO detect this
        lazy_problem().set_line(8)
        .set_text("A complex expression 'not (a + b > c and b + c > a and c + a > b)' used repeatedly "
                  "(on lines 8, 18, 27). Extract it to a local variable or create a helper function.")
    ]),
    ("014024-pythagorean.py", [
        lazy_problem().set_line(19)
        .set_text("A complex expression 'sqrt(a**2 + b**2)' used repeatedly (on lines 19, 20). Extract it to a "
                  "local variable or create a helper function."),
        lazy_problem().set_line(21)
        .set_text("A complex expression 'a + b + c > result' used repeatedly (on lines 21, 27, 33). Extract it to a "
                  "local variable or create a helper function."),
        lazy_problem().set_line(25),
        lazy_problem().set_line(31),
    ]),
    ("042643-edge_detection.py", [])
])
def test_duplicate_exprs(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=duplicate-exprs")],
        expected_output
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("013145-coins.py", [
        lazy_problem().set_line(5).set_end_line(7)
        .set_text("Duplicate blocks starting on lines 5 and 10. Extract the code to a helper function."),
        lazy_problem().set_line(10).set_end_line(12),
        lazy_problem().set_line(15).set_end_line(17),
        lazy_problem().set_line(20).set_end_line(22),
    ]),
    ("024508-cards.py", [
        lazy_problem().set_line(61).set_end_line(66),
    ])
])
def test_duplicate_blocks(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=duplicate-blocks")],
        expected_output
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("010666-coins.py", [
        lazy_problem().set_line(13).set_end_line(27)
        .set_text("Duplicate sequence of 5 repetitions of 3 lines of code. Use a loop to avoid this."),
    ]),
    ("014212-coins.py", [
        lazy_problem().set_line(8).set_end_line(19)
        .set_text("Duplicate sequence of 6 repetitions of 2 lines of code. Use a loop to avoid this."),
    ]),
    ("051796-gps.py", []),
    ("hw36211.py", []),
    ("ut156_9802_29_10.py", [
        lazy_problem().set_line(7).set_end_line(14)
        .set_text("Duplicate sequence of 4 repetitions of 2 lines of code. Use a loop to avoid this."),
        lazy_problem().set_line(15).set_end_line(24),
        lazy_problem().set_line(25).set_end_line(36),
        lazy_problem().set_line(37).set_end_line(48),
    ]),
])
def test_duplicate_sequence(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=duplicate-sequence")],
        expected_output
    )



from typing import List, Callable, Tuple, Dict, Set, Any
from edulint.linting.problem import ProblemJson, Problem
from edulint.linting.process_handler import ProcessHandler
from edulint.linting.overrides import get_overriders
from edulint.linting.tweakers import get_tweakers, Tweakers
from edulint.config.config import Config
from edulint.options import Option, ImmutableT
from edulint.linters import Linter
from functools import partial
import sys
import json
import os


def get_proper_path(path: str) -> str:
    return os.path.abspath(path) if os.path.isabs(path) else os.path.relpath(path)


def flake8_to_problem(raw: ProblemJson) -> Problem:
    assert isinstance(raw["filename"], str), f'got {type(raw["filename"])} for filename'
    assert isinstance(raw["line_number"], int), f'got {type(raw["line_number"])} for line_number'
    assert isinstance(
        raw["column_number"], int
    ), f'got {type(raw["column_number"])} for column_number'
    assert isinstance(raw["code"], str), f'got {type(raw["code"])} for code'
    assert isinstance(raw["text"], str), f'got {type(raw["text"])} for text'

    return Problem(
        Linter.FLAKE8,
        get_proper_path(raw["filename"]),
        raw["line_number"],
        raw["column_number"],
        raw["code"],
        raw["text"],
    )


def pylint_to_problem(filenames: List[str], raw: ProblemJson) -> Problem:
    assert isinstance(raw["path"], str), f'got {type(raw["path"])} for path'
    assert isinstance(raw["line"], int), f'got {type(raw["line"])} for line'
    assert isinstance(raw["column"], int), f'got {type(raw["column"])} for column'
    assert isinstance(raw["message-id"], str), f'got {type(raw["message-id"])} for message-id'
    assert isinstance(raw["message"], str), f'got {type(raw["message"])} for message'
    assert (
        isinstance(raw["endLine"], int) or raw["endLine"] is None
    ), f'got {type(raw["endLine"])} for endLine'
    assert (
        isinstance(raw["endColumn"], int) or raw["endColumn"] is None
    ), f'got {type(raw["endColumn"])} for endColumn'
    assert isinstance(raw["symbol"], str), f'get {type(raw["symbol"])} for symbol'

    def get_used_filename(path: str) -> str:
        for filename in filenames:
            if os.path.abspath(filename) == os.path.abspath(path):
                return filename
        assert False, "unreachable"

    return Problem(
        Linter.PYLINT,
        get_proper_path(get_used_filename(raw["path"])),
        raw["line"],
        raw["column"],
        raw["message-id"],
        raw["message"],
        raw["endLine"],
        raw["endColumn"],
        raw["symbol"],
    )


def lint_any(
    linter: Linter,
    filenames: List[str],
    linter_args: List[str],
    config_arg: ImmutableT,
    result_getter: Callable[[Any], Any],
    out_to_problem: Callable[[ProblemJson], Problem],
) -> List[Problem]:
    command = [sys.executable, "-m", str(linter)] + linter_args + list(config_arg) + filenames  # type: ignore
    return_code, outs, errs = ProcessHandler.run(command, timeout=1000)

    if ProcessHandler.is_status_code_by_timeout(return_code):
        print(f"edulint: {linter} was likely killed by timeout", file=sys.stderr)
        raise TimeoutError(f"Timeout from {linter}")

    print(errs, file=sys.stderr, end="")
    if (linter == Linter.FLAKE8 and return_code not in (0, 1)) or (
        linter == Linter.PYLINT and return_code == 32
    ):
        print(f"edulint: {linter} exited with {return_code}", file=sys.stderr)
        exit(return_code)

    if not outs:
        return []

    result = result_getter(json.loads(outs))
    return list(map(out_to_problem, result))


def lint_flake8(filenames: List[str], config: Config) -> List[Problem]:
    flake8_args = ["--format=json"]
    return lint_any(
        Linter.FLAKE8,
        filenames,
        flake8_args,
        config[Option.FLAKE8],
        lambda r: [problem for problems in r.values() for problem in problems],
        flake8_to_problem,
    )


def lint_pylint(filenames: List[str], config: Config) -> List[Problem]:
    pylint_args = ["--output-format=json"]
    return lint_any(
        Linter.PYLINT,
        filenames,
        pylint_args,
        config[Option.PYLINT],
        lambda r: r,
        partial(pylint_to_problem, filenames),
    )


def apply_overrides(problems: List[Problem], overriders: Dict[str, Set[str]]) -> List[Problem]:
    codes_on_lines: Dict[int, Set[str]] = {
        line: set() for line in set([problem.line for problem in problems])
    }

    for problem in problems:
        codes_on_lines[problem.line].add(problem.code)

    result = []
    for problem in problems:
        o = overriders.get(problem.code, set())
        if not (o & codes_on_lines[problem.line]):
            result.append(problem)

    return result


def apply_tweaks(problems: List[Problem], tweakers: Tweakers, config: Config) -> List[Problem]:
    result = []
    for problem in problems:
        tweaker = tweakers.get((problem.source, problem.code))
        if tweaker:
            if tweaker.should_keep(
                problem, [arg for arg in config if arg.option in tweaker.used_options]
            ):
                problem.text = tweaker.get_reword(problem)
                result.append(problem)
        else:
            result.append(problem)
    return result


def lint_one(filename: str, config: Config) -> List[Problem]:
    return lint([filename], config)


def sort(filenames: List[str], problems: List[Problem]) -> List[Problem]:
    indices = {get_proper_path(fn): i for i, fn in enumerate(filenames)}
    problems.sort(key=lambda problem: (indices[problem.path], problem.line, problem.column))
    return problems


def lint(filenames: List[str], config: Config) -> List[Problem]:
    result = ([] if config[Option.NO_FLAKE8] else lint_flake8(filenames, config)) + lint_pylint(
        filenames, config
    )
    result = apply_overrides(result, get_overriders())
    result = apply_tweaks(result, get_tweakers(), config)
    return sort(filenames, result)


def lint_many(partition: List[Tuple[List[str], Config]]) -> List[Problem]:
    return [problem for filenames, config in partition for problem in lint(filenames, config)]



from edulint.config.arg import ProcessedArg, UnprocessedArg, ImmutableArg
from edulint.linters import Linter
from edulint.options import (
    UnionT,
    ImmutableT,
    Option,
    DEFAULT_CONFIG,
    BASE_CONFIG,
    TakesVal,
    OptionParse,
    get_option_parses,
    get_name_to_option,
)
from edulint.config.file_config import load_toml_file
from edulint.config.config_translations import (
    get_config_translations,
    get_ib111_translations,
    Translation,
)
from typing import Dict, List, Optional, Tuple, Iterator, Any
from dataclasses import dataclass
from argparse import Namespace
import re
import sys
import shlex


@dataclass(frozen=True)
class Config:
    config: Tuple[ImmutableArg]

    @staticmethod
    def to_immutable(v: UnionT) -> ImmutableT:
        return v if not isinstance(v, list) else tuple(v)

    def __init__(
        self,
        config: Optional[List[ProcessedArg]] = None,
        option_parses: Dict[Option, OptionParse] = get_option_parses(),
    ) -> None:
        config = config if config is not None else []
        wip_config: List[Optional[ImmutableArg]] = [None for _ in Option]

        for arg in config:
            assert wip_config[int(arg.option)] is None
            wip_config[int(arg.option)] = ImmutableArg(arg.option, self.to_immutable(arg.val))

        object.__setattr__(
            self,
            "config",
            tuple(
                [
                    arg
                    if arg is not None
                    else ImmutableArg(o, self.to_immutable(option_parses[o].default))
                    for o, arg in zip(Option, wip_config)
                ]
            ),
        )

    def __str__(self) -> str:
        return f"Config({', '.join(arg.option.name + '=' + str(arg.val) for arg in self.config)})"

    def __getitem__(self, option: Option) -> ImmutableT:
        return self.config[int(option)].val

    def __contains__(self, option: Option) -> bool:
        return self[option] is not None

    def __iter__(self) -> Iterator[ImmutableArg]:
        return filter(lambda x: x is not None, self.config.__iter__())


def extract_args(filename: str) -> List[str]:
    edulint_re = re.compile(r"\s*#[\s#]*edulint:\s*", re.IGNORECASE)
    ib111_re = re.compile(r"\s*from\s+ib111\s+import\s+week_(\d+)", re.IGNORECASE)

    result: List[str] = []
    with open(filename, encoding="utf-8") as f:
        for i, line in enumerate(f):
            line = line.strip()

            edmatch = edulint_re.match(line)
            if edmatch:
                raw_args = line[edmatch.end() :]
                result.extend(shlex.split(raw_args))

            ibmatch = ib111_re.match(line)
            if ibmatch:
                result.append(f"{Option.IB111_WEEK.to_name()}={ibmatch.group(1)}")

    return result


def parse_option(
    option_parses: Dict[Option, OptionParse],
    name_to_option: Dict[str, Option],
    name: str,
    val: Optional[str],
) -> Optional[Option]:
    option = name_to_option.get(name)

    if option is None:
        print(f"edulint: unrecognized option {name}", file=sys.stderr)
    else:
        option_parse = option_parses[option]
        if option_parse.takes_val == TakesVal.YES and val is None:
            print(
                f"edulint: option {name} takes an argument but none was supplied",
                file=sys.stderr,
            )
        elif option_parse.takes_val == TakesVal.NO and val is not None:
            print(
                f"edulint: option {name} takes no argument but {val} was supplied",
                file=sys.stderr,
            )
        else:
            return option
    return None


def parse_args(
    args: List[str], option_parses: Dict[Option, OptionParse]
) -> Tuple[str, List[UnprocessedArg]]:
    name_to_option = get_name_to_option(option_parses)

    def get_name_val(arg: str) -> Tuple[str, Optional[str]]:
        if "=" in arg:
            name, val = arg.split("=", 1)
            return name, val
        return arg, None

    result: List[UnprocessedArg] = []
    config_path = DEFAULT_CONFIG
    for arg in args:
        name, val = get_name_val(arg)
        option = parse_option(option_parses, name_to_option, name, val)
        if option is not None:
            result.append(UnprocessedArg(option, val))
        if option == Option.CONFIG and val is not None:
            config_path = val

    return config_path, result


def parse_config_file(
    path: str, option_parses: Dict[Option, OptionParse]
) -> Optional[List[UnprocessedArg]]:
    def print_invalid_type_message(option: Option, val: Any) -> None:
        print(f"edulint: invalid value type {type(val)} of value {val} for option {Option.CONFIG}")

    def parse_base_config(config_dict: Dict[str, Any]) -> Optional[List[UnprocessedArg]]:
        rec_config = config_dict.get(Option.CONFIG.to_name(), BASE_CONFIG)
        if not isinstance(rec_config, str):
            print_invalid_type_message(Option.CONFIG, rec_config)
            rec_config = BASE_CONFIG
        return parse_config_file(rec_config, option_parses)

    def val_to_str(option: Option, val: Any) -> Optional[str]:
        if isinstance(val, str):
            return val
        if isinstance(val, list):
            return ",".join(val)
        if isinstance(val, tuple):
            key, value = val
            value = val_to_str(option, value)
            return f"--{key}={value}" if value is not None else None
        print_invalid_type_message(option, val)
        return None

    config_dict = load_toml_file(path)
    if config_dict is None:
        return None

    result = parse_base_config(config_dict) if path != BASE_CONFIG else []
    if result is None:
        return None

    name_to_option = get_name_to_option(option_parses)
    for name, val in config_dict.items():
        option = parse_option(option_parses, name_to_option, name, val)
        if option is None or option == Option.CONFIG:  # config is handled as the first option
            continue

        if not isinstance(val, dict):
            to_process = [val]
        elif option.to_name() not in {linter.to_name() for linter in Linter}:
            print_invalid_type_message(option, val)
            continue
        else:
            to_process = list(val.items())

        for val in to_process:
            str_val = val_to_str(option, val)
            if str_val is not None:
                result.append(UnprocessedArg(option, str_val))

    return result


def fill_in_val(arg: UnprocessedArg, translation: List[str]) -> List[str]:
    result = []
    for t in translation:
        if "<val>" in t:
            assert isinstance(arg.val, str)
            result.append(t.replace("<val>", arg.val))
        else:
            result.append(t)
    return result


def combine_and_translate(
    args: List[UnprocessedArg],
    option_parses: Dict[Option, OptionParse],
    config_translations: Dict[Option, Translation],
    ib111_translations: List[Translation],
) -> Config:
    def combine(option_vals: List[UnionT], option: Option, val: Optional[str]) -> None:
        parse = option_parses[option]
        old_val = option_vals[int(option)]
        option_vals[int(option)] = parse.combine(old_val, parse.convert(val))

    def apply_translation(option_vals: List[UnionT], translated: Translation) -> None:
        translated_option = translated.for_linter.to_option()
        for val in translated.vals:
            combine(option_vals, translated_option, val)

    option_vals = [option_parses[o].default for o in Option]

    for arg in args:
        combine(option_vals, arg.option, arg.val)

        translated = config_translations.get(arg.option)
        if translated is not None and option_vals[int(arg.option)] not in (False, None):
            apply_translation(option_vals, translated)

    ib111_week = option_vals[int(Option.IB111_WEEK)]
    if ib111_week is not None:
        assert isinstance(ib111_week, int)
        if 0 <= ib111_week < len(ib111_translations):
            apply_translation(option_vals, ib111_translations[ib111_week])
        else:
            print(
                f"edulint: option {Option.IB111_WEEK.to_name()} has value {ib111_week} which is invalid;"
                f"allowed values are 0 to {len(ib111_translations)}",
                file=sys.stderr,
            )

    return Config([ProcessedArg(o, v) for o, v in zip(Option, option_vals) if v is not None])


def get_config_one(
    filename: str,
    cmd_args: List[str],
    option_parses: Dict[Option, OptionParse] = get_option_parses(),
    config_translations: Dict[Option, Translation] = get_config_translations(),
    ib111_translation: List[Translation] = get_ib111_translations(),
) -> Optional[Config]:
    configs = get_config_many(
        [filename], cmd_args, option_parses, config_translations, ib111_translation
    )
    if len(configs) == 0:
        return None
    return configs[0][1]


def get_config_many(
    filenames: List[str],
    cmd_args_raw: List[str],
    option_parses: Dict[Option, OptionParse] = get_option_parses(),
    config_translations: Dict[Option, Translation] = get_config_translations(),
    ib111_translation: List[Translation] = get_ib111_translations(),
) -> List[Tuple[List[str], Config]]:
    def partition(
        configs: List[Tuple[str, List[UnprocessedArg]]]
    ) -> List[Tuple[List[str], Tuple[str, List[UnprocessedArg]]]]:
        immutable_configs = [(f, tuple(c)) for f, c in configs]

        dedup_configs = list(set(immutable_configs))
        indices = [dedup_configs.index(config) for config in immutable_configs]
        partitioned: List[List[str]] = [[] for _ in dedup_configs]

        for i, filename in enumerate(filenames):
            partitioned[indices[i]].append(filename)

        return list(zip(partitioned, [(f, list(c)) for f, c in dedup_configs]))

    cmd_config_path, cmd_args = parse_args(cmd_args_raw, option_parses)
    cmd_sets_config = any(arg.option == Option.CONFIG for arg in cmd_args)

    config_from_files = [
        parse_args(extract_args(filename), option_parses) for filename in filenames
    ]

    config_paths = {cmd_config_path} if cmd_sets_config else {c for c, _ in config_from_files}
    config_files_args = {
        config_path: parse_config_file(config_path, option_parses) for config_path in config_paths
    }

    result: List[Tuple[List[str], Config]] = []
    for files, (linted_config_path, linted_file_args) in partition(config_from_files):
        config_path = cmd_config_path if cmd_sets_config else linted_config_path
        config_file_args = config_files_args[config_path]
        if config_file_args is None:
            continue
        config = combine_and_translate(
            config_file_args + linted_file_args + cmd_args,
            option_parses,
            config_translations,
            ib111_translation,
        )
        result.append((files, config))
    return result


def get_cmd_args(args: Namespace) -> List[str]:
    return [s for arg in args.options for s in shlex.split(arg)]



import pytest
from edulint.linters import Linter
from edulint.config.arg import UnprocessedArg, ProcessedArg
from edulint.options import Option, TakesVal, Type, Combine, OptionParse, get_option_parses, DEFAULT_CONFIG
from edulint.config.config import Config, extract_args, parse_args, parse_config_file, combine_and_translate
from edulint.config.config_translations import get_config_translations, Translation
from edulint.linting.tweakers import get_tweakers
from typing import List, Set, Dict
from pathlib import Path


@pytest.fixture
def all_advertised_options() -> List[Option]:
    return list(get_option_parses().keys())


@pytest.fixture
def advertised_options(all_advertised_options: List[Option]) -> Set[Option]:
    return set(all_advertised_options)


@pytest.fixture
def always_managed_options() -> Set[Option]:
    return set((Option.CONFIG, Option.PYLINT, Option.FLAKE8, Option.IB111_WEEK, Option.NO_FLAKE8))


@pytest.fixture
def managed_translations_options() -> List[Option]:
    return list(get_config_translations().keys())


@pytest.fixture
def managed_tweaker_options() -> List[Set[Option]]:
    return list(tweaker.used_options for tweaker in get_tweakers().values())


@pytest.fixture
def managed_options(
        always_managed_options: Set[Option],
        managed_translations_options: List[Option],
        managed_tweaker_options: List[Set[Option]]
) -> Set[Option]:
    return always_managed_options \
        | set(managed_translations_options) \
        | set(opt for options in managed_tweaker_options for opt in options)


def issubset(subset, superset):
    diff = subset - superset
    assert not diff


def test_no_duplicate_advertised_options(all_advertised_options, advertised_options):
    issubset(set(all_advertised_options), advertised_options)


def test_all_advertised_options_managed(advertised_options: List[Option], managed_options: Set[Option]):
    issubset(advertised_options, managed_options)


def test_all_managed_options_are_advertised(advertised_options: List[Option], managed_options: Set[Option]) -> None:
    issubset(managed_options, advertised_options)


def mock_contents(mocker, contents) -> None:
    mocked_file = mocker.mock_open(read_data=contents)
    mocker.patch("builtins.open", mocked_file)


@pytest.mark.parametrize("contents,args", [
    ("", []),
    ("print(\"Hello world\")", []),
    ("# edulint: xxx", ["xxx"]),
    ("# edulint xxx", []),
    ("# edulint: pylint=xxx", ["pylint=xxx"]),
    ("# edulint: flake8=xxx", ["flake8=xxx"]),
    ("# edulint: xxx yyy", ["xxx", "yyy"]),
    ("# edulint: xxx\n# edulint: xxx", ["xxx", "xxx"]),
    ("# edulint: xxx\n# edulint: yyy", ["xxx", "yyy"]),
    ("# edulint: pylint=xxx\n# edulint: pylint=yyy", ["pylint=xxx", "pylint=yyy"]),
    ("# edulint: xxx\n# edulint: pylint=yyy", ["xxx", "pylint=yyy"]),
    ("# edulint: flake8=xxx\n# edulint: pylint=yyy", ["flake8=xxx", "pylint=yyy"]),
    ("\n\n# edulint: xxx", ["xxx"]),
    ("some code\n # edulint: xxx\ncode mentioning edulint a bit\n# edulint: yyy",
     ["xxx", "yyy"]),
    ("           #        edulint:      xxx        ", ["xxx"]),
    ("           #        edulint:      xxx  yyy      ", ["xxx", "yyy"]),
    ("# edulint: pylint=\"xxx       yyy\"", ["pylint=xxx       yyy"]),
    (
        "# edulint: pylint=xxx yyy zzz flake8=aaa bbb pylint= flake8=\"xxx yyy\" ",
        ["pylint=xxx", "yyy", "zzz", "flake8=aaa", "bbb", "pylint=", "flake8=xxx yyy"]
    ),
    ("# edulint: pylint=--enable=missing-module-docstring", ["pylint=--enable=missing-module-docstring"]),
    ("from ib111 import week_12", ["ib111-week=12"]),
    ("from ib111 import week_02  # noqa", ["ib111-week=02"]),
])
def test_extract_args(mocker, contents, args):
    mock_contents(mocker, contents)
    assert extract_args("foo") == args


@pytest.fixture
def options() -> Dict[Option, OptionParse]:
    return {
        Option.PYTHON_SPECIFIC: OptionParse(Option.PYTHON_SPECIFIC, "", TakesVal.NO, False, Type.BOOL, Combine.REPLACE),
        Option.FLAKE8: OptionParse(Option.FLAKE8, "", TakesVal.YES, [], Type.STR, Combine.APPEND),
        Option.IB111_WEEK: OptionParse(Option.IB111_WEEK, "", TakesVal.YES, None, Type.INT, Combine.REPLACE),
        Option.CONFIG: OptionParse(Option.CONFIG, "", TakesVal.YES, DEFAULT_CONFIG, Type.STR, Combine.REPLACE)
    }


@pytest.mark.parametrize("raw,parsed", [
    (["python-specific"], ("default", [UnprocessedArg(Option.PYTHON_SPECIFIC, None)])),
    (["python-spec"], ("default", [UnprocessedArg(Option.PYTHON_SPECIFIC, None)])),
    (["flake8=foo"], ("default", [UnprocessedArg(Option.FLAKE8, "foo")])),
    (["flake8="], ("default", [UnprocessedArg(Option.FLAKE8, "")])),
    (["python-specific", "flake8=foo"], ("default", [
       UnprocessedArg(Option.PYTHON_SPECIFIC, None), UnprocessedArg(Option.FLAKE8, "foo")
    ])),
    (["flake8=--enable=xxx"], ("default", [UnprocessedArg(Option.FLAKE8, "--enable=xxx")])),
    (["ib111-week=02"], ("default", [UnprocessedArg(Option.IB111_WEEK, "02")])),
    (["ib111-week=12", "ib111-week=02"], ("default", [
       UnprocessedArg(Option.IB111_WEEK, "12"), UnprocessedArg(Option.IB111_WEEK, "02")
    ])),
    (["config=empty"], ("empty", [UnprocessedArg(Option.CONFIG, "empty")]))
])
def test_parse_args(raw: List[str], options: Dict[Option, OptionParse], parsed: List[UnprocessedArg]) -> None:
    assert parse_args(raw, options) == parsed


def packaged_config_files():
    packaged_config_dir = Path(__file__).parent / ".." / "edulint" / "config" / "files"
    return [c.stem for c in packaged_config_dir.iterdir() if c.suffix == ".toml"]


@pytest.mark.parametrize("config_name", packaged_config_files())
def test_packaged_configs_parse(config_name: str):
    option_parses = get_option_parses()
    print(config_name)
    assert parse_config_file(config_name, option_parses) is not None


@pytest.fixture
def config_translations() -> Dict[Option, Translation]:
    return {
        Option.ENHANCEMENT: Translation(
            Linter.PYLINT,
            ["aaa"]
        ),
        Option.PYTHON_SPECIFIC: Translation(
            Linter.PYLINT,
            ["bbb"]
        ),
        Option.ALLOWED_ONECHAR_NAMES: Translation(
            Linter.PYLINT,
            ["ccc"]
        )
    }


@pytest.fixture
def ib111_translations() -> List[Translation]:
    return [
        Translation(Linter.PYLINT, ["iii"]),
        Translation(Linter.PYLINT, ["jjj"]),
        Translation(Linter.PYLINT, ["kkk"]),
    ]


@pytest.mark.parametrize("args,config", [
    (
        [UnprocessedArg(Option.ENHANCEMENT, None)],
        Config([ProcessedArg(Option.ENHANCEMENT, True), ProcessedArg(Option.PYLINT, ["aaa"])])
    ),
    (
        [UnprocessedArg(Option.ENHANCEMENT, None), UnprocessedArg(Option.PYLINT, "zzz")],
        Config([ProcessedArg(Option.ENHANCEMENT, True), ProcessedArg(Option.PYLINT, ["aaa", "zzz"])])
    ),
    (
        [UnprocessedArg(Option.PYLINT, "zzz"), UnprocessedArg(Option.ENHANCEMENT, None)],
        Config([ProcessedArg(Option.PYLINT, ["zzz", "aaa"]), ProcessedArg(Option.ENHANCEMENT, True)])
    ),
    (
        [UnprocessedArg(Option.FLAKE8, "zzz"), UnprocessedArg(Option.ENHANCEMENT, None)],
        Config([
            ProcessedArg(Option.FLAKE8, ["zzz"]),
            ProcessedArg(Option.ENHANCEMENT, True),
            ProcessedArg(Option.PYLINT, ["aaa"])
        ])
    ),
    (
        [UnprocessedArg(Option.ALLOWED_ONECHAR_NAMES, "n")],
        Config([ProcessedArg(Option.ALLOWED_ONECHAR_NAMES, "n"), ProcessedArg(Option.PYLINT, ["ccc"])])
    ),
    (
        [UnprocessedArg(Option.IB111_WEEK, "02")],
        Config([ProcessedArg(Option.IB111_WEEK, 2), ProcessedArg(Option.PYLINT, ["kkk"])])
    ),
    (
        [UnprocessedArg(Option.IB111_WEEK, "12"), UnprocessedArg(Option.IB111_WEEK, "02")],
        Config([ProcessedArg(Option.IB111_WEEK, 2), ProcessedArg(Option.PYLINT, ["kkk"])])
    ),
    (
        [UnprocessedArg(Option.IB111_WEEK, "02"), UnprocessedArg(Option.PYLINT, "aaa")],
        Config([ProcessedArg(Option.IB111_WEEK, 2), ProcessedArg(Option.PYLINT, ["aaa", "kkk"])])
    ),
    (
        [UnprocessedArg(Option.IB111_WEEK, "02"), UnprocessedArg(Option.ENHANCEMENT, None)],
        Config([
            ProcessedArg(Option.IB111_WEEK, 2),
            ProcessedArg(Option.ENHANCEMENT, True),
            ProcessedArg(Option.PYLINT, ["aaa", "kkk"])
        ])
    ),
])
def test_combine_and_translate_translates(
        args: List[UnprocessedArg],
        config_translations: Dict[Option, Translation],
        ib111_translations: List[Translation],
        config: Config) -> None:

    result = combine_and_translate(args, get_option_parses(), config_translations, ib111_translations)
    assert result.config == config.config



from pathlib import Path
import string
import os
from typing import Dict, Any, Optional
import sys

import tomli
import requests


ALLOWED_FILENAME_LETTERS = string.ascii_letters + string.digits + "-_"
ALLOW_UNRESTRICTED_LOCAL_PATHS = True
ALLOW_HTTP_S_PATHS = True


class EduLintConfigFileException(Exception):
    pass


class ConfigFileAccessMethodNotAllowedException(EduLintConfigFileException):
    pass


class InvalidConfigFile(EduLintConfigFileException):
    pass


def load_toml_file(filename_or_url: str) -> Optional[Dict[str, Any]]:
    try:
        file_content = _load_file_from_uri(filename_or_url)
    except Exception as e:
        print(
            f"edulint: Error locating config file '{filename_or_url}': {e}",
            file=sys.stderr,
        )
        return None

    try:
        file_toml = tomli.loads(file_content)
        return file_toml
    except tomli.TOMLDecodeError as e:
        print(
            f"edulint: Invalid TOML config file '{filename_or_url}': {e}",
            file=sys.stderr,
        )
        return None


def _load_file_from_uri(path_or_url: str) -> str:
    if path_or_url.startswith(("http://", "https://")):
        return _load_external_config_file(path_or_url)

    if _only_acceptable_chars(path_or_url):
        return _load_packaged_config_file(path_or_url)

    return _load_local_config_file(path_or_url, "found locally")


def _only_acceptable_chars(filepath: str) -> bool:
    return all([x in ALLOWED_FILENAME_LETTERS for x in filepath])


def _load_packaged_config_file(filename: str) -> str:
    assert _only_acceptable_chars(filename)

    relative_path = os.path.join(os.path.dirname(__file__), "files", filename + ".toml")
    return _load_local_config_file(relative_path, "packaged", is_path_safe=True)


def _load_local_config_file(filepath: str, message: str, is_path_safe: bool = False) -> str:
    if not is_path_safe and not ALLOW_UNRESTRICTED_LOCAL_PATHS:
        raise ConfigFileAccessMethodNotAllowedException(
            "Arbitrary local filepaths are not enabled."
        )

    # Doing the test in two steps should prevent possible exception during the if test.
    if not (Path(filepath).exists() and Path(filepath).is_file()):
        raise FileNotFoundError(f"Configuration file '{filepath}' not {message}.")

    with open(filepath, encoding="utf8") as f:
        return f.read()


def _load_external_config_file(url: str) -> str:
    if not ALLOW_HTTP_S_PATHS:
        raise ConfigFileAccessMethodNotAllowedException(
            "Loading of external configs using HTTP/HTTPS is disallowed in EduLint's configuration."
        )

    resp = requests.get(url)
    if resp.status_code != 200:
        raise FileNotFoundError(
            f"Request for external config '{url}' failed with status code {resp.status_code}."
        )

    return resp.content.decode()



import pytest
from edulint.options import Option
from edulint.config.arg import Arg
from edulint.linting.problem import Problem
from utils import lazy_problem, apply_and_lint, create_apply_and_lint
from typing import List


def _test_simplify_if(lines: List[str], expected_output: List[Problem], code: str) -> None:
    create_apply_and_lint(
        lines,
        [Arg(Option.PYLINT, f"--enable={code}")],
        [p.set_code(code) for p in expected_output]
    )


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def yyy(x):",
        "    if x:",
        "        return True",
        "    else:",
        "        return False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x'")
    ]),
    ([
        "def xxx(x):",
        "    if x:",
        "        return False",
        "    else:",
        "        return True"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return <negated x>'"),
    ]),
    ([
        "def xxx(x):",
        "    if x:",
        "        return True",
        "    return False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x'"),
    ]),
    ([
        "def xxx(x):",
        "    if x:",
        "        return False",
        "    return True"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return <negated x>'"),
    ]),
])
def test_simplify_if_statement_single_var_custom(lines: List[str], expected_output: List[Problem]) -> None:
    _test_simplify_if(lines, expected_output, "R6201")


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def xxx(x, y):",
        "    if x:",
        "        return True",
        "    return y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x or y'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        return False",
        "    return y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return <negated x> and y'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        return y",
        "    return False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x and y'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        return y",
        "    return True"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return <negated x> or y'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        return True",
        "    else:",
        "        return y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x or y'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        return False",
        "    else:",
        "        return y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return <negated x> and y'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        return y",
        "    else:",
        "        return False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x and y'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        return y",
        "    else:",
        "        return True"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return <negated x> or y'")
    ]),
])
def test_simplify_if_statement_two_vars_custom(lines: List[str], expected_output: List[Problem]) -> None:
    _test_simplify_if(lines, expected_output, "R6202")


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def xxx(x, y, z):",
        "    if x and z:",
        "        return y",
        "    return False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x and z and y'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x or z:",
        "        return y",
        "    return False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return (x or z) and y'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x:",
        "        return y and z",
        "    return False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x and y and z'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x:",
        "        return y or z",
        "    return False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x and (y or z)'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x and z:",
        "        return True",
        "    return y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return (x and z) or y'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x and z:",
        "        return False",
        "    return y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return <negated (x and z)> and y'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x:",
        "        return False",
        "    return y or z"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return <negated x> and (y or z)'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x or z:",
        "        return y",
        "    return False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return (x or z) and y'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x and z:",
        "        return y",
        "    return True"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return <negated (x and z)> or y'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x and z:",
        "        return True",
        "    else:",
        "        return y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return (x and z) or y'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x or z:",
        "        return True",
        "    else:",
        "        return y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x or z or y'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x:",
        "        return True",
        "    else:",
        "        return y and z"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x or (y and z)'")
    ]),
])
def test_simplify_if_statement_three_vars_custom(lines: List[str], expected_output: List[Problem]) -> None:
    _test_simplify_if(lines, expected_output, "R6202")


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def xxx(x, y):",
        "    if x:",
        "        if y:",
        "            return 0",
        "    return 1"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be merged with the nested one to 'if x and y:'")
    ]),
])
def test_simplify_if_statement_nested_custom(lines: List[str], expected_output: List[Problem]) -> None:
    _test_simplify_if(lines, expected_output, "R6207")


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def xxx(x, y):",
        "    if x:",
        "        return 0",
        "    if y:",
        "        return 0",
        "    return 1"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be merged with the following one to 'if x or y:'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        return True",
        "    if y:",
        "        return False",
        "    return True"
    ], [
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x and y:",
        "        return True",
        "    if z:",
        "        return True",
        "    return True"
    ], [
    ]),
])
def test_simplify_if_statement_multiconds_custom(lines: List[str], expected_output: List[Problem]) -> None:
    _test_simplify_if(lines, expected_output, "R6208")


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def is_right(a, b, c):",
        "    if c ** 2 == a ** 2 + b ** 2 or a ** 2 == c ** 2 + b ** 2 or \\",
        "       b ** 2 == a ** 2 + c ** 2:",
        "        triangle_is_righ = True",
        "    else:",
        "        triangle_is_righ = False",
        "    return triangle_is_righ"
    ], [
        lazy_problem().set_line(2)
        .set_text("The conditional assignment can be replace with 'triangle_is_righ = c**2 == a**2 "
                  "+ b**2 or a**2 == c**2 + b**2 or b**2 == a**2 + c**2'")
    ]),
])
def test_simplify_if_assigning_statement_custom(lines: List[str], expected_output: List[Problem]) -> None:
    _test_simplify_if(lines, expected_output, "R6203")


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def xxx(x, y):",
        "    if x:",
        "        a = True",
        "    else:",
        "        a = y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The conditional assignment can be replace with 'a = x or y'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        a = False",
        "    else:",
        "        a = y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The conditional assignment can be replace with 'a = <negated x> and y'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        a = y",
        "    else:",
        "        a = False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The conditional assignment can be replace with 'a = x and y'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        a = y",
        "    else:",
        "        a = True"
    ], [
        lazy_problem().set_line(2)
        .set_text("The conditional assignment can be replace with 'a = <negated x> or y'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x:",
        "        a = y and z",
        "    else:",
        "        a = True"
    ], [
        lazy_problem().set_line(2)
        .set_text("The conditional assignment can be replace with 'a = <negated x> or (y and z)'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x:",
        "        a = y or z",
        "    else:",
        "        a = True"
    ], [
        lazy_problem().set_line(2)
        .set_text("The conditional assignment can be replace with 'a = <negated x> or y or z'")
    ]),
])
def test_simplify_if_assigning_statement_conj_custom(lines: List[str], expected_output: List[Problem]) -> None:
    _test_simplify_if(lines, expected_output, "R6210")


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "report = []",
        "which = 0",
        "report[which] = True if report[which] > \\",
        "    report[which] else False"
    ], [
        lazy_problem().set_line(3)
        .set_text("The if expression can be replaced with 'report[which] > report[which]'")
    ]),
    ([
        "report = []",
        "which = 0",
        "report[which], x = True if report[which] > report[which] else False, 0"
    ], [
        lazy_problem().set_line(3)
        .set_text("The if expression can be replaced with 'report[which] > report[which]'")
    ]),
    ([
        "report = []",
        "which = 0",
        "report[which], x = True if report[which] > report[which] else False, \\",
        "    False if report[which] <= report[which] else True"
    ], [
        lazy_problem().set_line(3)
        .set_text("The if expression can be replaced with 'report[which] > report[which]'"),
        lazy_problem().set_line(4)
        .set_text("The if expression can be replaced with '<negated report[which] <= report[which]>'")
    ]),
])
def test_simplify_if_expression_custom(lines: List[str], expected_output: List[Problem]) -> None:
    _test_simplify_if(lines, expected_output, "R6204")


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def xxx(x, y):",
        "    r = True if x else y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if expression can be replaced with 'x or y'")
    ]),
    ([
        "def xxx(x, y):",
        "    r = False if x else y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if expression can be replaced with '<negated x> and y'")
    ]),
    ([
        "def xxx(x, y):",
        "    r = y if x else False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if expression can be replaced with 'x and y'")
    ]),
    ([
        "def xxx(x, y):",
        "    r = y if x else True"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if expression can be replaced with '<negated x> or y'")
    ]),
])
def test_simplify_if_expression_conj_custom(lines: List[str], expected_output: List[Problem]) -> None:
    _test_simplify_if(lines, expected_output, "R6209")


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def xxx(x, y):",
        "    if x:",
        "        pass",
        "    else:",
        "        return y"
    ], [
        lazy_problem().set_line(2)
        .set_text("Use 'if <negated x>: <else body>' instead of 'pass'")
    ]),
])
def test_simplify_if_pass_custom(lines: List[str], expected_output: List[Problem]) -> None:
    _test_simplify_if(lines, expected_output, "R6205")


@pytest.mark.parametrize("filename,expected_output", [
    ("015080-p4_geometry.py", [
        lazy_problem().set_code("R6207").set_line(14)
        .set_text("The if statement can be merged with the nested one to 'if len(sides) != len(set(sides))"
                  " and len(set(sides)) <= 3 and sides[1] == sides[2]:'"),
        lazy_problem().set_code("R6201").set_line(21)
        .set_text("The if statement can be replaced with 'return side_c == sides[2]'"),
        lazy_problem().set_code("R6201").set_line(32)
        .set_text("The if statement can be replaced with 'return a == b & a == c'"),
        lazy_problem().set_code("R6201").set_line(47)
        .set_text("The if statement can be replaced with 'return <negated duplicate(sides) < max(sides)>'"),
    ]),
    ("z202817-zkouska.py", [
        lazy_problem().set_code("R6208").set_line(76)
        .set_text("The if statement can be merged with the following one to "
                  "'if word == '' or word[len(word) - 1] == '.':'"),
        lazy_problem().set_code("R6205").set_line(82),
    ]),
    ("014823-p4_geometry.py", [
            lazy_problem().set_code("R6206").set_line(17)
            .set_text("Both branches should return a value explicitly (one returns implicit None)"),
            lazy_problem().set_code("R6206").set_line(25),
            lazy_problem().set_code("R6206").set_line(33),
            lazy_problem().set_code("R6201").set_line(43),
            lazy_problem().set_code("R6201").set_line(55),
    ]),
    ("024423-p5_credit.py", []),
    ("044669-p3_person_id.py", [lazy_problem().set_code("R6206").set_line(30)])
])
def test_simplify_if_files(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=simplifiable-if")],
        expected_output
    )



from edulint.config.config import get_config_many, get_cmd_args
from edulint.linting.problem import Problem
from edulint.linting.linting import lint_many, sort
from typing import List, Optional
import argparse
import os
import sys


def setup_argparse() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Lints provided code.")
    parser.add_argument(
        "-o", "--option", dest="options", default=[], action="append", help="possible option"
    )
    parser.add_argument(
        "files_or_dirs",
        metavar="FILE-OR-DIRECTORY",
        nargs="+",
        help="the file(s) or directory(ies) to lint",
    )
    parser.add_argument("--json", action="store_true", help="should output problems in json format")
    return parser.parse_args()


def extract_files(files_or_dirs: List[str]) -> List[str]:
    def extract_files_rec(
        prefix: Optional[str], files_or_dirs: List[str], result: List[str]
    ) -> List[str]:
        for file_or_dir in files_or_dirs:
            full_path = os.path.join(prefix, file_or_dir) if prefix is not None else file_or_dir
            if os.path.isdir(full_path):
                extract_files_rec(full_path, os.listdir(full_path), result)
            elif os.path.splitext(full_path)[1] == ".py":
                result.append(full_path)
        return result

    return extract_files_rec(None, files_or_dirs, [])


def main() -> int:
    args = setup_argparse()
    cmd_args = get_cmd_args(args)

    files = extract_files(args.files_or_dirs)
    file_configs = get_config_many(files, cmd_args)

    try:
        result = sort(files, lint_many(file_configs))
    except TimeoutError as e:
        print(f"edulint: {e}", file=sys.stderr)
        sys.exit(1)

    if args.json:
        print(Problem.schema().dumps(result, indent=2, many=True))  # type: ignore
    else:
        prev_problem = None
        for problem in result:
            if len(files) > 1 and (prev_problem is None or prev_problem.path != problem.path):
                print(f"****************** {os.path.basename(problem.path)}")
                prev_problem = problem

            print(problem)
    return 0
